import os.path

CONFIG_JSON = os.path.expanduser('~/.dataset_sh_config.json')
STORAGE_BASE = os.path.expanduser('~/dataset_sh/storage')
ALIAS_FILE = os.path.expanduser('~/.dataset.sh.host-alias.json')
CENTRAL_HOST = 'https://remote.dataset.sh'
